#ifndef LIST_H
#define LIST_H

#include <cstddef>

/**
* List can be used to collect pointers to elements of type T.
*/
template <class T>
class List {
  public:
    static const int INITIALSIZE = 30;
    static const int REALLOC_SIZESTEP = 10;
  public:
    List<T>(int p_InitialSize = INITIALSIZE);
    bool  addElement(T* p_Element); 
    bool  removeElement(T* p_Element);
    void  clear();
    int   getSize();
    T*    getElement(int p_Index);
    int   getCapacity();
    bool  getRallocSucceeded();
  protected:
      T**   m_Elements;
      int   m_ElementCount = 0;
      int   m_Capacity = 0;
      bool  m_ReallocSucceeded = true;

};


template <class T>
List<T>::List(int p_InitialSize) {
  m_Capacity = p_InitialSize;
  m_Elements = new T*[m_Capacity];
}

template <class T>
int List<T>::getCapacity() {
  return m_Capacity;
}
template <class T>
bool List<T>::getRallocSucceeded() {
  return m_ReallocSucceeded;
}

template <class T>
bool List<T>::addElement(T* p_Element) {
  if (p_Element == NULL) return false;
  T* ptrElement = NULL;

  //checkif p_Element already exists in m_Elements
  for (int intElement = 0; intElement < m_ElementCount; intElement++) {
    ptrElement = m_Elements[intElement];
    if (ptrElement == p_Element) {
      break;
    }
    ptrElement = NULL;
  }
  if (ptrElement != NULL) {
    //p_Input already exists in m_Elements
    m_ReallocSucceeded = false;
    return true;
  }


  if (m_ElementCount >= m_Capacity) {
    //try to realloc m_Elements
    T** newArray;
    int intNewCapacity;
    intNewCapacity = m_Capacity + REALLOC_SIZESTEP;
    newArray = (T**)realloc(m_Elements, sizeof(T*) * intNewCapacity);
    if (newArray != NULL) {
      m_Elements = newArray;
      for (int intElement = m_Capacity; intElement < intNewCapacity; intElement++) {
        m_Elements[intElement] = NULL;
      }
      m_Capacity = intNewCapacity;
    }
    else {
      //raalloc fails
      return false;
    }
  }
  

  //add at the end (m_ElementCount)
  m_Elements[m_ElementCount] = p_Element;
  m_ElementCount++;

  return true;
}

template <class T>
bool List<T>::removeElement(T* p_Element) {
  if (p_Element == NULL) return false;
  int intElement;
  T* ptrElement = NULL;
  if (m_ElementCount <= 0) {
    return false;
  }
  for (intElement = 0; intElement < m_ElementCount; intElement++) {
    ptrElement = m_Elements[intElement];
    if (ptrElement == p_Element) {
      break;
    }
    else {
      ptrElement = NULL;
    }
  }
  if (ptrElement == NULL) {
    //p_Input does not exist in s_Instnces
    return false;
  }
  //move down inputs above intElement 
  for (intElement = intElement; intElement < m_ElementCount-1; intElement++) {
    ptrElement = m_Elements[intElement+1];
    m_Elements[intElement] = ptrElement;
  }
  //set last to NULL
  m_Elements[m_ElementCount-1] = NULL;
  m_ElementCount--;
  return true;
} 
template <class T>
void List<T>::clear() {
  for (int intElement = 0; intElement < m_ElementCount; intElement++) {
    m_Elements[intElement] = NULL;
  }
  m_ElementCount = 0;
}

template <class T>
int List<T>::getSize() {
  return m_ElementCount;
}
template <class T>
T* List<T>::getElement(int p_Index) {
  if (p_Index < 0) return NULL;
  if (p_Index >= m_ElementCount) return NULL;
  return m_Elements[p_Index];
}

#endif