#ifndef TIMEHANDLER_H
#define TIMEHANDLER_H

#include "delay.h"
#include "List.h"

/**
* Interface used by the TimerHandler to notify all listeners of the changed time.
* This can be used for debouncing of inputs or avoiding using delay().
*/
class TimerListener {
  public:
    /** TimerHandler notifies all added TimerListener about time is changed by calling eventTimer with the actual time (micros). 
    */
    virtual void eventTimer(uint64_t p_ActualMicros) = 0;
};

/**
* The TimerHandler can be used to iplement non blockimg actions and pseudo parallel Actions or for debouncing inputs.* The TimerHandler handles a list of TimerListeners.
* Whenever refreshTime() is called every TimerListener will be notified by calling eventTimer with the actual time (millis).
* Perhaps calling of refeshTime may be done by a timer interrupt or by manually calling in the sketch's loop function.
*/
class TimerHandler {
  protected:
    //static list of all added TimerListeners 
    static List<TimerListener>* s_List_TimerListeners;
    static unsigned long  s_LastMicros; //millis, of the last call of refreshTime
    static unsigned long  s_TimerOverflows; //count the overflow of getMicros()
    static uint64_t       s_ActualTime_Micros;
  public:
    //adds a TimerListener to the static list
    static bool addTimerListener(TimerListener* p_TimerListener);
    //removes a Timerlistener from the static list
    static bool removeTimerListener(TimerListener* p_TimerListener);
    //returns the amount of listet TimerListeners
    static int  getTimerListenerCount();
    /**
    * Returns the listet TimerListener at position: p_Index.
    * If p_Index is outside of the range, NULL is returned.
    */
    static TimerListener* getTimerListener(int p_Index);
    /**
    * Should be called up at regular intervals.
    * All listet TimerListeners are notified of the current time (millis).
    * Perhaps this may called from timer interrupt or manually in the sketch's loop() function.
    */ 
    static void refreshTime();
    /** return sthe actual time in micro seconds
    */
    static uint64_t getTime_ActualMicros();
    /** Returns the difference between p_TimeMicros and the actual time in micro seconds
     * @return time difference between actual time and p_TimeMicros in micro seconds
    */
    static uint64_t getTimeDiff_Micros(uint64_t p_TimeMicros);
    /** return sthe actual time in milliseconds
    */
    static uint64_t getTime_ActualMillis();
    /** Returns the difference between p_TimeMillis and the actual time in milliseconds
     * @return time difference between actual time and p_TimeMillis in milli seconds
    */
    static uint64_t getTimeDiff_Millis(uint64_t p_TimeMillis);
};

#endif