#include "delay.h"
#include <stdint.h>
#include "LoraHandler.h"

//-------------------- Lora Message ------------------

String LoraMessage::VALUE_SEPERATOR = ":";

LoraMessage::LoraMessage() {
  int intPrecodeLength;
  intPrecodeLength = LENGTH_PRECODELENGTH + LENGTH_SENDERID + LENGTH_SENDERTYPE;
  m_Bytes[0] = intPrecodeLength & 0xFF;
  setSenderID(0);      //should be changed by calling setSenderID(..)
  setSenderType(' ');  //maybe changed by calling setSenderType(char), if several types of senders exists
}
LoraMessage::~LoraMessage() {
  
}
int LoraMessage::getMaxLength() {
  return RH_RF95_MAX_MESSAGE_LEN;
}
char* LoraMessage::getBytes() {
  return m_Bytes;
}

void LoraMessage::setLength(int p_Length) {
  m_Length = p_Length;
}
int LoraMessage::getLength() {
  return m_Length;
}
unsigned int LoraMessage::getPrecodeLength() {
  return m_Bytes[0] & 0xFF;
}
void LoraMessage::setSenderID(uint32_t p_SenderID) {
  m_Bytes[LENGTH_PRECODELENGTH + 0] = p_SenderID & 0xFF;
  m_Bytes[LENGTH_PRECODELENGTH + 1] = (p_SenderID >> 8) & 0xFF;
  m_Bytes[LENGTH_PRECODELENGTH + 2] = (p_SenderID >> 16) & 0xFF;
  m_Bytes[LENGTH_PRECODELENGTH + 3] = (p_SenderID >> 24) & 0xFF;
}

void LoraMessage::setSenderType(char p_SenderType) {
  m_Bytes[LENGTH_PRECODELENGTH + LENGTH_SENDERID] = p_SenderType;
}
void LoraMessage::setMessage(String p_Message, String p_Value) {
  String stringMessageAll = "";
  if (p_Message != NULL) {
    stringMessageAll += p_Message;
  }
  if (p_Value != NULL) {
    stringMessageAll += VALUE_SEPERATOR + p_Value;
  }

  char* ptrMessage;
  ptrMessage = m_Bytes;
  ptrMessage += getPrecodeLength();

  if (stringMessageAll.length() + 1 >= RH_RF95_MAX_MESSAGE_LEN - getPrecodeLength()) {
    stringMessageAll = stringMessageAll.substring(0, RH_RF95_MAX_MESSAGE_LEN - getPrecodeLength() - 1);
  }
  stringMessageAll.toCharArray(ptrMessage, RH_RF95_MAX_MESSAGE_LEN - getPrecodeLength());
  m_Length = getPrecodeLength() + stringMessageAll.length() + 1;
}
uint32_t LoraMessage::getSenderID() {
  uint32_t uintID = 0;
  uintID = m_Bytes[LENGTH_PRECODELENGTH + 0];
  uintID += m_Bytes[LENGTH_PRECODELENGTH + 1] << 8;
  uintID += m_Bytes[LENGTH_PRECODELENGTH + 2] << 16;
  uintID += m_Bytes[LENGTH_PRECODELENGTH + 3] << 24;
  return uintID;
}
String LoraMessage::getSenderID_Hex() {
  char string[30];
  ultoa( getSenderID(), string, 16);
  return string;
}
char LoraMessage::getSenderType() {
  return m_Bytes[LENGTH_PRECODELENGTH + LENGTH_SENDERID];
}
String LoraMessage::getMessage() {
  char* ptrMessage = m_Bytes;
  ptrMessage += getPrecodeLength();
  String stringMessage(ptrMessage);
  return stringMessage;
}
String LoraMessage::getMessageWithoutValue() {
  String stringMessage= getMessage();
  int intIndexSeperator;
  intIndexSeperator = stringMessage.lastIndexOf(VALUE_SEPERATOR);
  if (intIndexSeperator >= 0) {
    stringMessage = stringMessage.substring(0, intIndexSeperator);
  }
  return stringMessage;
}
String LoraMessage::getValueString() {
  String stringValue= getMessage();
  int intIndexSeperator;
  intIndexSeperator = stringValue.lastIndexOf(VALUE_SEPERATOR);
  if (intIndexSeperator >= 0) {
    stringValue = stringValue.substring(intIndexSeperator+1, stringValue.length());
  }
  return stringValue;
}
double LoraMessage::getValue() {
  double doubleValue = -1;
  String stringValue = getValueString();
  if (stringValue.length() > 0) {
    doubleValue = stringValue.toDouble();
  }
  return doubleValue;
}

String LoraMessage::getString() {
  String stringResult;
//  stringResult += getMessage();
  stringResult = "LoraSender(";
  stringResult += getSenderID();
  stringResult += ")_";
  stringResult += getMessageWithoutValue();
  stringResult += "_";
  if (getValue() == 0) {
    stringResult += "Down";
  }
  else {
    stringResult += "Up";
  }
//  stringResult += getValueString();

  return stringResult;
}
void LoraMessage::calculateMessageSerial() {
  String stringMessage;
  int intLength;
  stringMessage = getString();
  intLength = stringMessage.length() + 1;
  if (intLength > MESSAGESERIAL_LENGTH) {
    intLength = MESSAGESERIAL_LENGTH;
  }
  stringMessage.toCharArray(m_MessageSerial_Bytes, intLength);
  m_MessageSerial_Bytes[intLength-1] = 0x00;
  m_MessageSerial_Length = intLength-1;
}
char* LoraMessage::getMessageSerial_Buffer() {
  return m_MessageSerial_Bytes;
}
size_t LoraMessage::getMessageSerial_Length() {
  return m_MessageSerial_Length;
}

void LoraMessage::printToSerial() {
  Serial.print("LoraMessage Length: "); 
  Serial.print(getLength());
  Serial.print(" PrecodeLength: "); 
  Serial.print(getPrecodeLength());
  Serial.print(" SenderID: "); 
  Serial.print(getSenderID());
  Serial.print(" SenderType: "); 
  Serial.print(getSenderType());
  Serial.print(" RSSI: "); 
  Serial.println(Lora::getastRSSI());
  
  Serial.print("Message: "); 
  Serial.print(getMessageWithoutValue());
  Serial.print(" Value: ");
  Serial.println(getValue());

}



//-------------------- Lora ------------------


RH_RF95 Lora::rf95(RFM95_CS, RFM95_INT);
bool Lora::initCalled = false;

uint32_t* Lora::s_SenderIDs = new uint32_t[MAX_SENDERIDS];
char* Lora::s_Sender_StartCode = new char[MAX_SENDERIDS];
int Lora::s_SenderIDCount = 0;

int Lora::s_EEPROM_HeaderSize = 5; //byte[0-3] = 'L''o''r''a' byte[4] = Amount of SenderIDs, other bytes contains the sender IDs
int Lora::s_EEPROM_Array_MaxLength = s_EEPROM_HeaderSize + MAX_SENDERIDS * LoraMessage::LENGTH_SENDERID;
char* Lora::s_EEPROM_Array = new char[s_EEPROM_Array_MaxLength]; //byte[0-3] = 'L''o''r''a' byte[4] = Amount of SenderIDs, other bytes contains the sender IDs
StorageHandler* Lora::s_StorageHandler = NULL;

Output* Lora::s_Output_LED_IDS_CLeared = NULL;
Action_Blink* Lora::s_Action_Blink_Learning = NULL;

bool Lora::s_LearningActive = false;
bool Lora::s_Learning_ChangedSenderID = false;;


bool Lora::init(Output* p_Output_LED_Response) {
  if (initCalled == true) return true; //init is prior called successfully
  s_Output_LED_IDS_CLeared = p_Output_LED_Response;
  s_Action_Blink_Learning = new Action_Blink(s_Output_LED_IDS_CLeared, 50, 50, -1, Action_Blink::MODE_START_INLOW_STOP_INHIGH);


  clearSenderIDs(); //clear the sender IDs

  int intFailureCounter = 0;
  bool boolFailure = false;

  pinMode(RFM95_RST, OUTPUT);
  digitalWrite(RFM95_RST, HIGH);

  // manual reset
  digitalWrite(RFM95_RST, LOW);
  delay(10);
  digitalWrite(RFM95_RST, HIGH);
  delay(10);

  while (!rf95.init()) {
    delay(1);
    intFailureCounter++;
    if (intFailureCounter > 1000) {
      boolFailure = true;
      break;
    }
  }
  if (boolFailure) {
    return false;
  }
 
  // Defaults after init are 434.0MHz, modulation GFSK_Rb250Fd250, +13dbM
  if (!rf95.setFrequency(RF95_FREQ)) {
    boolFailure = true;
    return false;
  }

  // Defaults after init are 434.0MHz, 13dBm, Bw = 125 kHz, Cr = 4/5, Sf = 128chips/symbol, CRC on

  // The default transmitter power is 13dBm, using PA_BOOST.
  // If you are using RFM95/96/97/98 modules which uses the PA_BOOST transmitter pin, then
  // you can set transmitter powers from 5 to 23 dBm:
  rf95.setTxPower(20, false);
  
  initCalled = true;
  return true;
}


//--- SenderIDs ---
void Lora::clearSenderIDs() {
  for (int intIndex = 0; intIndex < MAX_SENDERIDS; intIndex++) {
    s_SenderIDs[intIndex] = 0;
    s_Sender_StartCode[intIndex] = DEFAULT_STARTCODE;
  }
  if (s_SenderIDCount != 0) {
    s_SenderIDCount = 0;
    if (s_LearningActive == true) {
      s_Learning_ChangedSenderID = true; //set flag for saving changed SenderIDs
    }
  }
}
bool Lora::addSenderID(uint32_t p_SenderID) {
  uint32_t intSenderID;
  if (s_SenderIDCount >= MAX_SENDERIDS) {
    return false;
  }
  for (int intIndex = 0; intIndex < s_SenderIDCount; intIndex++) {
    intSenderID = s_SenderIDs[intIndex];
    if (intSenderID == p_SenderID) {
      return false;
    }
  }

  s_SenderIDs[s_SenderIDCount] = p_SenderID;
  s_Sender_StartCode[s_SenderIDCount] = DEFAULT_STARTCODE;
  s_SenderIDCount++;
  if (s_LearningActive == true) {
    //set learning SenderIDs changed flag
    s_Learning_ChangedSenderID = true;  //set flag for saving changed SenderIDs
  }
  return true;  
}
uint32_t Lora::getSenderID(int p_Index) {
  if (p_Index <0) return 0;
  if (p_Index >= s_SenderIDCount) return 0;
  return s_SenderIDs[p_Index];
}
bool Lora::containsSenderID(uint32_t p_SenderID) {
  if (p_SenderID == 0) return false;
  uint32_t intSenderID;
  for (int intIndex = 0; intIndex < s_SenderIDCount; intIndex++) {
    intSenderID = s_SenderIDs[intIndex];
    if (intSenderID == p_SenderID) {
      return true;
    }
  }
  return false;
}
int Lora::getSenderID_Index(u_int32_t p_SenderID) {
  if (p_SenderID == 0) return -1;
  uint32_t intSenderID;
  for (int intIndex = 0; intIndex < s_SenderIDCount; intIndex++) {
    intSenderID = s_SenderIDs[intIndex];
    if (intSenderID == p_SenderID) {
      return intIndex;
    }
  }
  return -1;
}

char Lora::getSenderID_StartCode(uint32_t p_SenderID) {
  int intIndex;
  intIndex = getSenderID_Index(p_SenderID);
  if (intIndex < 0) return DEFAULT_STARTCODE;
  return s_Sender_StartCode[intIndex];
}
bool Lora::setSenderID_StartCode(uint32_t p_SenderID, char p_StartCode) {
  int intIndex;
  intIndex = getSenderID_Index(p_SenderID);
  if (intIndex < 0) return false;
  if (p_StartCode == s_Sender_StartCode[intIndex]) {
    return false;
  }
  s_Sender_StartCode[intIndex] = p_StartCode;
  return true;
}

int Lora::getSenderIDCount() {
  return s_SenderIDCount;
}


//--- Learning ---
void Lora::setLearningActive(bool p_LearningActive) {
  s_LearningActive = p_LearningActive;
  if (s_LearningActive == true) {
    s_Learning_ChangedSenderID = false;
    s_Action_Blink_Learning->run(LOW);
  }
  else {
    //s_LearningActive == false
    s_Action_Blink_Learning->run(HIGH);
    if (s_Learning_ChangedSenderID == true) {
      //save SenderIDs only when something has changed
      Lora::saveSenderIDs();
      s_Learning_ChangedSenderID = false; 
    }
  }
}
bool Lora::getLearningActive() {
  return s_LearningActive;
}


//--- Send ---
bool Lora::send(LoraMessage* p_LoraMessage) {
  if (p_LoraMessage == NULL) return false;
  if (p_LoraMessage->getLength() <= 0) return false;
  //send message

  if (rf95.send((uint8_t *)(p_LoraMessage->getBytes()), (uint8_t)(p_LoraMessage->getLength())) == true) {
    rf95.waitPacketSent();
  }
  else {
    return false;
  }
  return true;
}

bool Lora::getAvailable() {
  return rf95.available();
}

int Lora::read(LoraMessage* p_LoraMessage) {
  int intIndex = -1;
  if (p_LoraMessage == NULL) return -1;
  if (rf95.available() == true) {
    uint8_t len = RH_RF95_MAX_MESSAGE_LEN;
    if (rf95.recv((uint8_t*)p_LoraMessage->getBytes(), &len)) {
      p_LoraMessage->setLength(len);
      intIndex = getSenderID_Index(p_LoraMessage->getSenderID());
      if (s_LearningActive || intIndex >= 0) {
        //only accept LoraMessage, where the sender id is known or learning is active
//        p_LoraMessage->calculateMessageSerial();
        return len;
      }
    }
    else {
      //could not recv message
    }
  }
  return -1;
}

int Lora::getastRSSI() {
  return rf95.lastRssi();
}



//--- Save/Read SenderIDs ----
void Lora::setStorageHandler(StorageHandler* p_StorageHandler) {
  s_StorageHandler = p_StorageHandler;
}

void Lora::saveSenderIDs() {

  if (s_StorageHandler == NULL) {
    return;
  }
  int intBytesWritten;
  int intBytesToWrite;
  int intSenderIDCount = 0;

  digitalWrite(RFM95_CS, HIGH); //switch off Lora

  //erase sector to allow write in this sector
  s_StorageHandler->eraseSector(0);

  fillEEPROM_Array();
  intSenderIDCount = s_EEPROM_Array[s_EEPROM_HeaderSize-1] & 0xFF;
  intBytesToWrite = s_EEPROM_HeaderSize + LoraMessage::LENGTH_SENDERID * intSenderIDCount;

  intBytesWritten = s_StorageHandler->writeByteArray(0, s_EEPROM_Array, 0, intBytesToWrite);
  
  if (intBytesWritten == intBytesToWrite) {  
    readSenderIDs();
  }

  digitalWrite(RFM95_CS, LOW); //switch on Lora

}
void Lora::readSenderIDs() {
  if (s_StorageHandler == NULL) {
    return;
  }

  digitalWrite(RFM95_CS, HIGH); //switch off Lora

  int intBytesReaden;
  uint8_t intSenderIDCount = 0;
  //read header "LORA" + SenderCount
  intBytesReaden = s_StorageHandler->readByteArray(0, s_EEPROM_Array, 0, s_EEPROM_HeaderSize);
  if (intBytesReaden != s_EEPROM_HeaderSize) {
    digitalWrite(RFM95_CS, LOW); //switch on Lora
    return;
  }
  intSenderIDCount = s_EEPROM_Array[s_EEPROM_HeaderSize-1] & 0xFF;
  if (intSenderIDCount > MAX_SENDERIDS) {
    intSenderIDCount = MAX_SENDERIDS;
  }

  int intAddress;
  int intOffset;
  int intBytesToRead;

  intAddress = s_EEPROM_HeaderSize;
  intOffset = s_EEPROM_HeaderSize;

  intBytesToRead = intSenderIDCount * LoraMessage::LENGTH_SENDERID;
  intBytesReaden = s_StorageHandler->readByteArray(intAddress, s_EEPROM_Array, intOffset, intBytesToRead);
  if (intBytesReaden != intBytesToRead) {
    digitalWrite(RFM95_CS, LOW); //switch on Lora
    return;
  }

  interpreteEEPROM_Array(); //SenderIDs will be cleared and new added from EEPROM

  digitalWrite(RFM95_CS, LOW); //switch on Lora

}


char* Lora::fillEEPROM_Array() {
  //Header = "Lora" + 1 byte for SenderIDCount
  s_EEPROM_Array[0] = 'L';
  s_EEPROM_Array[1] = 'o';
  s_EEPROM_Array[2] = 'r';
  s_EEPROM_Array[3] = 'a';

  s_EEPROM_Array[s_EEPROM_HeaderSize-1] = (uint8_t)getSenderIDCount() & 0xFF;


  uint32_t uintID;
  for (int intSendwerID = 0; intSendwerID < getSenderIDCount(); intSendwerID++) {
    uintID = getSenderID(intSendwerID);
    s_EEPROM_Array[s_EEPROM_HeaderSize + intSendwerID * 4 + 0] = uintID & 0xFF;
    s_EEPROM_Array[s_EEPROM_HeaderSize + intSendwerID * 4 + 1] = (uintID >> 8) & 0xFF;
    s_EEPROM_Array[s_EEPROM_HeaderSize + intSendwerID * 4 + 2] = (uintID >> 16) & 0xFF;
    s_EEPROM_Array[s_EEPROM_HeaderSize + intSendwerID * 4 + 3] = (uintID >> 24) & 0xFF;
  }
  return s_EEPROM_Array;
}

bool Lora::interpreteEEPROM_Array() {
  int intSenderIDCount;
//  if (sizeof(p_EEPROM_Array) != s_EEPROM_Array_MaxLength) return false;
  if (s_EEPROM_Array[0] != 'L') return false;
  if (s_EEPROM_Array[1] != 'o') return false;
  if (s_EEPROM_Array[2] != 'r') return false;
  if (s_EEPROM_Array[3] != 'a') return false;
  uint32_t uintID = 0;

  clearSenderIDs();

  intSenderIDCount = s_EEPROM_Array[s_EEPROM_HeaderSize-1] & 0xFF;
  for (int intSenderID = 0; intSenderID < intSenderIDCount; intSenderID++) {
    uintID = s_EEPROM_Array[s_EEPROM_HeaderSize + intSenderID * 4 + 0];
    uintID += s_EEPROM_Array[s_EEPROM_HeaderSize + intSenderID * 4 + 1] << 8;
    uintID += s_EEPROM_Array[s_EEPROM_HeaderSize + intSenderID * 4 + 2] << 16;
    uintID += s_EEPROM_Array[s_EEPROM_HeaderSize + intSenderID * 4 + 3] << 24;
    addSenderID(uintID); //will increase s_SenderIDCount
  }
  return true;
}



//-------------------- Action_Lora_Send ------------------

Action_Lora_Send::Action_Lora_Send(uint32_t p_SenderID, char p_SenderType, String p_Message) : Action() {
  m_LoraMessage = new LoraMessage();
  m_LoraMessage->setSenderID(p_SenderID);
  m_LoraMessage->setSenderType(p_SenderType);
  m_MessageWithoutValue = p_Message;
}
Action_Lora_Send::~Action_Lora_Send() {
  delete m_LoraMessage;
}

void Action_Lora_Send::setMessage(String p_Message) {
  m_MessageWithoutValue = p_Message;
}

void Action_Lora_Send::setValueDecimalPlaces(int p_DecimalPlaces) {
  m_ValueDecimalPlaces = p_DecimalPlaces;
}

bool Action_Lora_Send::run(double p_ValueIn) {
  String stringValue;
  if (m_ValueDecimalPlaces >= 0) {
    stringValue = String(p_ValueIn, m_ValueDecimalPlaces);
    stringValue.trim();
  }

  m_LoraMessage->setMessage(m_MessageWithoutValue, stringValue);

  return Lora::send(m_LoraMessage);
}

void Action_Lora_Send::eventTimer(uint64_t p_ActualMicros) {
}

//-------------------- Action_Lora_Received ------------------
Action_Lora_Received::Action_Lora_Received() : Action_Serial() {
  m_Action_HID = new Action_HID();
}
Action_Lora_Received::~Action_Lora_Received() {
  delete m_Action_HID;
}
void Action_Lora_Received::setInput_Lora(Input_Lora *p_Input_Lora) {
  m_Input_Lora = p_Input_Lora;
  Action_Serial::setAction_Serial_Interface(m_Input_Lora);
}
bool Action_Lora_Received::run(double p_ValueIn) {
  if (p_ValueIn == HIGH) return false;
  if (m_Input_Lora == NULL) return false;
  if (Serial) {
    //send serial messaage
    Action_Serial::run(p_ValueIn);
    return true;
  }
  //Send message as HID
  String stringMessage;
  LoraMessage* ptrLoraMessage;
  ptrLoraMessage = m_Input_Lora->getLoraMessage();
  if (ptrLoraMessage == NULL) return false;
  stringMessage = ptrLoraMessage->getMessage();
  if (stringMessage.length() > 0) {
      //set char for keypressed/ keyReleased: (char)(stringMessage[0]
      char charMessage;
      charMessage = (char)stringMessage[0];
      m_Action_HID->setKeyboardMouseEvent(charMessage, 0);
      m_Action_HID->run(m_Input_Lora->getLoraMessage()->getValue());
  }
  return true;
}

    
//-------------------- Action_LoraLearning ------------------

Action_LoraLearning::Action_LoraLearning(Output* p_Output_LED) : Action() {
  m_Output_ResponseClear = p_Output_LED;
  m_ActionStarted = false;
  m_LearningActive = false;
  m_ShouldRun = false;
}
Action_LoraLearning::~Action_LoraLearning() {
}

bool Action_LoraLearning::run(double p_Value_In) {
  if (p_Value_In == LOW) {
    m_ActionStarted = true;
    m_LearningActive = Lora::getLearningActive();
    if (m_LearningActive == false) {
      m_ShouldRun = true;
      m_Clear_IsDone = false;
    }
    else {
      m_ShouldRun = false;
    }
  }
  else {
    m_Clear_IsDone = true;
    if (m_Output_ResponseClear != NULL) m_Output_ResponseClear->writeValue(LOW);
  }
  return true;
}

void Action_LoraLearning::eventTimer(uint64_t p_ActualMicros) {
  if (m_ActionStarted == false) return; 
  if (m_ShouldRun == false) {
    if (m_LearningActive == true) {
      //stop
      Lora::setLearningActive(false);
      m_ActionStarted = false;
    }
    return;
  }
  //m_ShouldRun == true
  if (m_LearningActive == false) {
    //start
    m_StartTime = millis();
    Lora::setLearningActive(true); //swap learning
    m_LearningActive = true;
    return;
  }
  else {

  }
  if (m_Clear_IsDone == false) {
    if (TimerHandler::getTimeDiff_Millis(m_StartTime) > TIME_CLEARSERVERIDS_MS) {
      m_StartTime = millis();
      Lora::clearSenderIDs();
      m_Clear_IsDone = true;
      Lora::setLearningActive(false);
      m_LearningActive = false;
      if (m_Output_ResponseClear != NULL) m_Output_ResponseClear->writeValue(HIGH);
      m_ShouldRun = false;
    }
  }
}


//-------------------- Input_Lora ------------------

Input_Lora::Input_Lora() {
  m_LoraMessage_Received = new LoraMessage();
}
Input_Lora::~Input_Lora() {

}
String Input_Lora::getClassName() {
  return "Input_Lora";
}

double Input_Lora::readValue() {
  if (Lora::read(m_LoraMessage_Received) <= 0) {
    //return pseudo value HIGH because there is no message readen
    return HIGH;
  }

  changeMessageStartCode();
  m_LoraMessage_Received->calculateMessageSerial();

  if (m_SendResponse == true) {
      
     //send the same message as response
    Lora::send(m_LoraMessage_Received);
  }

  setValueChanged(true);  //setValueChanged to true because a new message arrived.
                            //and the last time readValue was called, a new message may also have arrived. -> the last returnd vaue was also LOW
  //return pseudo value LOW because a new message is received
  return LOW;
  
}

void Input_Lora::changeMessageStartCode() {
  if (m_LoraMessage_Received == NULL) return;
  uint32_t uintSenderID;
  int intIndex;
  char charStartCode;
  String stringMessage;
  char charMessage;
  uintSenderID = m_LoraMessage_Received->getSenderID();
  intIndex = Lora::getSenderID_Index(uintSenderID);
  if (intIndex < 0) {
    //Sender id not found
    return;
  }
  charStartCode = Lora::getSenderID_StartCode(uintSenderID);
  if (charStartCode == Lora::DEFAULT_STARTCODE) {
    //Startcose is default -> do nothing
    return;
  }
  //change the message (startCode diff)
  int intStartCodeDiff;
  intStartCodeDiff = charStartCode - Lora::DEFAULT_STARTCODE;

  stringMessage = m_LoraMessage_Received->getMessageWithoutValue();
  if (stringMessage.length() <= 0) {
    return;
  }

  charMessage = stringMessage[0];
  charMessage += intStartCodeDiff;
  stringMessage = String(charMessage);
  m_LoraMessage_Received->setMessage(stringMessage, m_LoraMessage_Received->getValueString());
}

void Input_Lora::setSendResponse(bool p_SendResponse) {
  m_SendResponse = p_SendResponse;
}

LoraMessage* Input_Lora::getLoraMessage() {
  return m_LoraMessage_Received;
}

char* Input_Lora::getMessageSerial_Buffer() {
  if (m_LoraMessage_Received == NULL) {
    return NULL;
  }
  
  return m_LoraMessage_Received->getMessageSerial_Buffer();
}

size_t Input_Lora::getMessageSerial_Length() {
  if (m_LoraMessage_Received == NULL) {
    return 0;
  }
  return m_LoraMessage_Received->getMessageSerial_Length();
}


bool Input_Lora::runActions() {
  if (Lora::getLearningActive() == true) {
    //Lora is in learning mode -> don't run normal actions but add the received sender id
    if (m_LoraMessage_Received != NULL) {
      if (m_LoraMessage_Received->getValue() == HIGH) {
        //first LoraMessage received has the value LOW (opposite button is pressed)
        //second oraMessage received has the value HIGH (opposite button is released)
        //react when the value of the received lora message is HIGH
        Lora::addSenderID(m_LoraMessage_Received->getSenderID());
        Lora::setLearningActive(false);
      }
    }
    return false;
  }
  return Input::runActions();
  
}



String Action_Serial_ChangeStartCode::COMMAND_START = "lorasender";

Action_Serial_ChangeStartCode::Action_Serial_ChangeStartCode() {

}
Action_Serial_ChangeStartCode::~Action_Serial_ChangeStartCode() {

}
void Action_Serial_ChangeStartCode::setAction_Serial_InputInterface(Action_Serial_InputInterface* p_Action_Serial_InputInterface) {
  m_Action_Serial_InputInterface = p_Action_Serial_InputInterface;
}
bool Action_Serial_ChangeStartCode::run(double p_Value_In) {
  if (p_Value_In == HIGH) {
    return false;
  }
  if (m_Action_Serial_InputInterface == NULL) {
    return false;
  }
  size_t o_MessageLength;
  char* o_MessageBuffer;

  String stringCommandStart;
  String stringMessage = "";
  
  int intStartID;
  int intEndID;
  String stringSenderID;
  uint32_t uintSenderID;
  int intSenderID_Index;
  
  String stringStartCode;
  char charStartCode;


  o_MessageLength = m_Action_Serial_InputInterface->getMessageSerial_Length();

  if (o_MessageLength <= 0) {
    return false;
  }
  o_MessageBuffer = m_Action_Serial_InputInterface->getMessageSerial_Buffer();
  for (size_t intChar = 0; intChar < o_MessageLength; intChar++) {
    stringMessage += o_MessageBuffer[intChar];
  }
  stringMessage.trim();
  
  //Check Coomand start
  if (stringMessage.length() < COMMAND_START.length()) {
    return false;
  }
  stringCommandStart = stringMessage.substring(0, o_MessageLength);
  stringCommandStart.toLowerCase();
  if (stringCommandStart.startsWith(COMMAND_START) == false) {
    return false;
  }

  //find SenderID
  intStartID = stringMessage.indexOf('(');
  intEndID = stringMessage.indexOf(')');
  if (intStartID < 0 || intEndID < 0) {
    return false;
  }
  if (intStartID >= intEndID) {
    return false;
  }
  stringSenderID = stringMessage.substring(intStartID+1, intEndID);
  uintSenderID = (uint32_t)stringSenderID.toInt();
  if (uintSenderID == 0) {
    return false;
  }
  //check if SenderID is in list of SenderID;
  intSenderID_Index = Lora::getSenderID_Index(uintSenderID);
  if (intSenderID_Index < 0) {
    return false;
  }
  //check 
  stringStartCode = stringMessage.substring(intEndID+1, intEndID+3);
  if (stringStartCode.length() < 2) {
    return false;
  }
  if (stringStartCode.startsWith("_") == false) {
    return false;
  }
  stringStartCode = stringStartCode.substring(1,2);
  stringStartCode.toUpperCase();
  charStartCode = stringStartCode[0];
  if ('Z' +1 - charStartCode < 8) {
    return false;
  }
 
  Lora::setSenderID_StartCode(uintSenderID, charStartCode); //change the startcode
                                                            //receiver will interprete the received message:
                                                            // charCode + (charStartCode - Lora::DEFAULT_STARTCODE)

  return true;
}
void Action_Serial_ChangeStartCode::eventTimer(uint64_t p_ActualMicros) {

}
