#include "StorageHandler.h"

StorageHandler::StorageHandler(int p_Pin_CS, int p_PageSize, int p_AddressBitCount,
    uint32_t p_Clock, BitOrder p_BitOrder, uint8_t p_DataMode) {
    m_Pin_CS = p_Pin_CS;
    m_PageSize_Bytes = p_PageSize;
    m_AddressBitCount = p_AddressBitCount;
    
    //SPI EEPROM
    pinMode(m_Pin_CS, OUTPUT);
    digitalWrite(m_Pin_CS, HIGH);
    SPI.begin();
    
    m_SPISettings = new SPISettings(p_Clock, p_BitOrder, p_DataMode);


}
StorageHandler::~StorageHandler() {
    delete m_SPISettings;
}


void StorageHandler::eraseSector(uint64_t p_Address) {

        SPI.beginTransaction(*m_SPISettings);

        digitalWrite(m_Pin_CS, LOW); //switch on SPI
        SPI.transfer(SPI_WRITEENABLE);  //enable write
        digitalWrite(m_Pin_CS, HIGH); //switch off SPI
       
        digitalWrite(m_Pin_CS, LOW);
            SPI.transfer(SPI_ERASESECTOR);

            switch (m_AddressBitCount) {
                    case 32:
                        SPI.transfer((byte)(p_Address >> 24 & 0xFF));
                    case 24:
                        SPI.transfer((byte)(p_Address >> 16 & 0xFF));
                        case 16:
                        SPI.transfer((byte)(p_Address >> 8 & 0xFF));
                        default:
                        SPI.transfer ((byte)(p_Address & 0xFF));
                    }
        digitalWrite(m_Pin_CS, HIGH);
        delay(400);
            


        SPI.endTransaction();
}

int StorageHandler::writeByteArray(uint64_t p_Address, char p_ByteArray[], int p_Offset, int p_BytesToWrite) {
    int intBytesWritten = 0;
    uint64_t intPage;
    int intPageRestBytes = 0;
    int intBytesToWrite;


    intPage = p_Address / m_PageSize_Bytes;
    intPageRestBytes = (intPage+1) * m_PageSize_Bytes - p_Address;

    if (p_BytesToWrite > intPageRestBytes) {
        intBytesToWrite = intPageRestBytes;
    }
    else {
        intBytesToWrite = p_BytesToWrite;
    }   
        SPI.beginTransaction(*m_SPISettings);

        digitalWrite(m_Pin_CS, LOW); //switch on SPI
        SPI.transfer(SPI_WRITEENABLE);  //enable write
        digitalWrite(m_Pin_CS, HIGH); //switch off SPI
       
        while(p_BytesToWrite > 0) {
            digitalWrite(m_Pin_CS, LOW);
            SPI.transfer(SPI_WRITE);

            switch (m_AddressBitCount) {
                    case 32:
                        SPI.transfer((byte)(p_Address >> 24 & 0xFF));
                    case 24:
                        SPI.transfer((byte)(p_Address >> 16 & 0xFF));
                        case 16:
                        SPI.transfer((byte)(p_Address >> 8 & 0xFF));
                        default:
                        SPI.transfer ((byte)(p_Address & 0xFF));
                    }
            for (int intByte = 0; intByte < intBytesToWrite; intByte++) {
                //Serial.print("    WriteByte ");
                //Serial.println(p_ByteArray[p_Offset + intByte], HEX);
                SPI.transfer(p_ByteArray[p_Offset + intByte]);
            }
            digitalWrite(m_Pin_CS, HIGH);
            delay(5);
            
            intBytesWritten += intBytesToWrite;
            p_Address += intBytesToWrite;
            p_BytesToWrite -= intBytesToWrite;
            p_Offset += intBytesToWrite;

            if (p_BytesToWrite > m_PageSize_Bytes) {
                intBytesToWrite = m_PageSize_Bytes;
            }

        }

        SPI.endTransaction();

    return intBytesWritten;
}

int StorageHandler::readByteArray(uint64_t p_Address, char p_ByteArray[], int p_Offset, int p_BytesToRead) {
    int intBytesReaden = 0;
    int intBytesToRead = 0;

 
 
//    intPage = p_Address / m_PageSize_Bytes;
 //   intPageRestBytes = (intPage+1) * m_PageSize_Bytes - p_Address;

//    if (p_BytesToRead > intPageRestBytes) {
//        intBytesToRead = intPageRestBytes;
//    }
//    else {
        intBytesToRead = p_BytesToRead;
//    }    
        SPI.beginTransaction(*m_SPISettings);
       
//        while(p_BytesToRead > 0) {
            digitalWrite(m_Pin_CS, LOW);
            SPI.transfer(SPI_READ);

            switch (m_AddressBitCount) {
                    case 64:
                        SPI.transfer((byte)(p_Address >> 24 & 0xFF));
                    case 32:
                        SPI.transfer((byte)(p_Address >> 24 & 0xFF));
                    case 24:
                        SPI.transfer((byte)(p_Address >> 16 & 0xFF));
                    case 16:
                        SPI.transfer((byte)(p_Address >> 8 & 0xFF));
                    default:
                        SPI.transfer ((byte)(p_Address & 0xFF));
            }
            for (int intByte = 0; intByte < intBytesToRead; intByte++) {
                p_ByteArray[p_Offset + intByte] = SPI.transfer(0x00);
                //Serial.print("    Byte readen ");
                //Serial.println(p_ByteArray[p_Offset + intByte], HEX);
            }
            digitalWrite(m_Pin_CS, HIGH);
            
            intBytesReaden += intBytesToRead;
            p_Address += intBytesToRead;
            p_BytesToRead -= intBytesToRead;
            p_Offset += intBytesReaden;

//            if (p_BytesToRead > m_PageSize_Bytes) {
//                intBytesToRead = m_PageSize_Bytes;
//            }

 //       }


    SPI.endTransaction();


    return intBytesReaden;
}

