#include <Arduino.h>
#include <SPI.h>
#include "InOutputHandler.h"
#include "LoraHandler.h"
#include "IDHandler.h"
#include "StorageHandler.h"


#define Pin_Output_LED 13
#define Pin_Input_Learning 11
//#define Pin_Input_Learning 12
//StorageHandler has to be rewritten for specific Flash/EEprom chips
#define STORAGE_HANDLER_PIN_CS 19
#define STORAGE_HANDLER_PAGESIZE 256
#define STORAGE_HANDLER_ADDRESSBITCOUNT 24


//--------------------- MAIN --------------------

uint32_t  uintBoardID; //checksum of the serial number of this arduino hardware

Action_Blink* o_Action_Blink_Start;

Input_Lora* o_Input_Lora;
Action_Lora_Received* o_Action_Lora_Received;

Output_Digital* o_Output_Digital_LED;

Input_Digital* o_InputDigital_Learning;
Action_Blink* o_Action_Blink_LED;
Action_LoraLearning* o_Action_LoraLearning;

Input_Serial* o_InputSerial;
Action_Serial_ChangeStartCode* o_Action_Serial_ChangeStartCode;

void setup() {

  //Get the Board Serialnumber as a checksum
  //The whole SerialNumber may be to great for sending it with a lora message
  uintBoardID = SerialNumber::getCheckSum_32(); //we use the checksum of the board's serial number as Lora SenderID (4 byte)

  o_Output_Digital_LED = new Output_Digital(Pin_Output_LED);

  //--- Lora INIT ---
  if (Lora::init(o_Output_Digital_LED) == false) {
    //error Lora initialization
  }
  //Create a StorageHandler instance and add it to Lora
  //The StorageHandler has to be rewritten for specific Flash//EEPROM chips
  StorageHandler* ptrStorageHandler;
  ptrStorageHandler = new StorageHandler(STORAGE_HANDLER_PIN_CS, STORAGE_HANDLER_PAGESIZE, STORAGE_HANDLER_ADDRESSBITCOUNT);
  Lora::setStorageHandler(ptrStorageHandler);
  
  //Actions

  o_Action_Blink_LED = new Action_Blink(o_Output_Digital_LED, 200, 50, 1, Action_Blink::MODE_RUNONCE_INHIGH);
  o_Action_LoraLearning = new Action_LoraLearning(o_Output_Digital_LED);
  
  o_Action_Serial_ChangeStartCode = new Action_Serial_ChangeStartCode();

  o_Action_Lora_Received = new Action_Lora_Received();

   //--- Inputs ---
  
  o_InputDigital_Learning = new Input_Digital(Pin_Input_Learning);
  o_InputDigital_Learning->addAction(o_Action_LoraLearning);


  o_Input_Lora = new Input_Lora();
  o_Input_Lora->setDebounceTimeSpan(0);
  o_Input_Lora->setSendResponse(true);
  o_Input_Lora->addAction(o_Action_Blink_LED);
  
  o_Action_Lora_Received->setInput_Lora(o_Input_Lora);
  o_Input_Lora->addAction(o_Action_Lora_Received);
 

  o_InputSerial = new Input_Serial();
  o_InputSerial->addAction(o_Action_Serial_ChangeStartCode);
  o_Action_Serial_ChangeStartCode->setAction_Serial_InputInterface(o_InputSerial);

  //--- Lora read saved SenderIDs ---
  Lora::readSenderIDs();

  if (Lora::getSenderIDCount() <= 0) {
    //if no sender id is readen -> start Learning
    Lora::setLearningActive(true);
  }
  else {
    //Blink to show setup is completed
    o_Action_Blink_Start = new Action_Blink(o_Output_Digital_LED, 500,500,3,Action_Blink::MODE_RUNONCE_INLOW);
    o_Action_Blink_Start->run(LOW);
  }
}

/**
 * The loop is fixed, no need to change it
 */
void loop() {
 
  //Refresh TimerHandler, all TimerListeners will be notified
  TimerHandler::refreshTime(); //this is important for Actions and Input debouncing, eventTimer will be called

  
  Input* ptrInput;
  
  for (int intInput = 0; intInput < Input::getInputCount(); intInput++) {
    ptrInput = Input::getInput(intInput);
    if (ptrInput == NULL) continue;

    if (ptrInput->isInterruptable() == false) {
      //the input has no interrupt -> call checkInput (the next timerevent will call input's checkValueChanged)
      ptrInput->checkInput();
    }


    if (ptrInput->getValueChanged() == true) {
      //value of the input has changed
      ptrInput->runActions(); //run all actions of the input

      ptrInput->processed();  //has to be called when the changed value is handled - this will set Input's m_ValueChanged to false
    }
  }

  if (Input::getAllInterruptable() == true || Input::getInputCount() == 0) {
    //all inputs are interruptable or no inputs exist -> perhaps we can use deepSleep 
    //Actuall interruptables are not used 
    //and perhaps there is no need to use interruptles, because TimerHandler makes the interruptables unnecessary
//    Serial.println("All Inputs interruptable");
  }
 
  /*
  //To delete an Input and free memory use Input::deleteInput()  for excymple:
  if (m_Input_Digital_1 != NULL) {
    (Input_Digital*)Input::deleteInput(m_Input_Digital_1);
    m_Input_Digital_1 = NULL;
  }
  */

}

